package me.hammerle.code;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Stack;
import me.hammerle.math.Fraction;

public class Code implements Comparable<Code>
{
    private final int line;
    private final int subline;
    
    private final String function;
    private final int level;
    private final int pars;
    
    private final Object value;
    private int jump;
    
    public Code(String function, int level, int pars, int line, int subline, Object value, int jump)
    {
        this.function = function;
        this.level = level;
        this.pars = pars;
        this.line = line;
        this.subline = subline;
        this.value = value;
        this.jump = 0;
    }
    
    public Code(String function, int level, int pars, int line, int subline)
    {
        this(function.trim(), level, pars, line, subline, null, 0);
    }
    
    public Code(int level, int pars, int line, int subline, Object value)
    {
        this(null, level, pars, line, subline, value, 0);
    }
    
    public void executeFunction(Script sc, Stack<Object> stack)
    {
        if(value != null)
        {
            if(value.getClass() == Variable.class)
            {
                stack.push(sc.getVar(((Variable) value).getName()));
                return;
            }
            stack.push(value);
            return;
        }
        Object[] input = null;
        if(pars > 0)
        {
            input = new Object[pars];
            for(int i = 0; i < pars; i++)
            {
                input[i] = stack.pop();
            }
        }
        Object output = CodeParser.parseFunction(sc, function, input, jump);
        if(output != Void.TYPE)
        {
            stack.push(output);
        }
    }

    @Override
    public String toString() 
    {
        StringBuilder sb = new StringBuilder(">>>");
        for(int i = 1; i < level; i++)
        {
            sb.append(">>>");
        }
        sb.append(" (");
        sb.append(line);
        sb.append(") ");
        
        if(value != null)
        {
            sb.append("push ");
            sb.append(value.getClass());
            sb.append(" value = '");
            sb.append(value);
            sb.append("'");
            return sb.toString();
        }
        
        sb.append("function ");
        sb.append(function);
        sb.append(" (");
        for(int i = 0; i < pars; i++)
        {
            sb.append("X,");
        }
        if(pars > 0)
        {
            sb.deleteCharAt(sb.length() - 1);
        }
        sb.append(")");
        if(jump != 0)
        {
            sb.append(" (");
            sb.append(jump);
            sb.append(")");
        }
        return sb.toString();
    }
    
    @Override
    public int compareTo(Code o) 
    {
        int i = Integer.compare(line, o.line);
        if(i == 0)
        {
            return Integer.compare(o.subline, subline);
        }
        return i;
    }
    
    // -----------------------------------------------------------------------------------
    // Code-Builder
    // -----------------------------------------------------------------------------------
    
    private static int sublines = 0;
    
    private static int findEndOfLine(String code, int pos)
    {
        int length = code.length();
        char c;
        boolean text = false;
        while(pos < length)
        {
            c = code.charAt(pos);
            if(text && c != '"')
            {
                pos++;
                continue;
            }
            switch(c)
            {
                case ';':
                case '{':
                case '}':
                    return pos;
                case '"':
                    text = !text;
                    break;
            }
            pos++;
        }
        return -1;
    }
    
    public static Code[] generate(String code)
    {
        ArrayList<Code> list = new ArrayList<>();
        sublines = 0;
        int length = code.length();
        int level = 1;
        int pos = 0;
        int old;
        int line = 0;
        while(pos < length)
        {
            old = pos;
            pos = findEndOfLine(code, pos);
            switch(code.charAt(pos))
            {
                case '{':
                    line++;
                    splitFunctions(list, code.substring(old, pos).trim(), level, line);
                    level++;
                    break;
                case '}': 
                    level--;
                    break;
                case ';': 
                    line++;
                    splitFunctions(list, code.substring(old, pos).trim(), level, line);
                    break;
            }
            pos++;
        }
        Collections.sort(list);
        
        String function;
        int baseLevel;
        for(int i = 0; i < list.size(); i++)
        {
            function = list.get(i).function;
            if(function == null)
            {
                continue;
            }
            switch(function)
            {
                case "if":
                case "while":
                case "try":
                case "catch":
                    break;
                default:
                    continue;
            }
            baseLevel = list.get(i).level;
            for(int j = i + 1; j < list.size(); j++)
            {
                if(list.get(j).level == baseLevel)
                {
                    list.get(i).jump = j - i;
                    break;
                }
            }
            if(list.get(i).jump == 0)
            {
               list.get(i).jump = list.size() - i; 
            }
        }
        
        list.forEach(c -> System.out.println(c.toString()));
        return list.toArray(new Code[list.size()]);
    } 
    
    private static int findOpenBracket(String code, int pos)
    {
        int length = code.length();
        char c;
        boolean text = false;
        while(pos < length)
        {
            c = code.charAt(pos);
            if(text && c != '"')
            {
                pos++;
                continue;
            }
            switch(c)
            {
                case '(':
                    return pos;
                case '"':
                    text = !text;
                    break;
            }
            pos++;
        }
        return length;
    }
    
    private static int findClosingBracket(String code, int pos)
    {
        int length = code.length();
        char c;
        boolean text = false;
        int brackets = 0;
        while(pos < length)
        {
            c = code.charAt(pos);
            if(text && c != '"')
            {
                pos++;
                continue;
            }
            switch(c)
            {
                case '(':
                    brackets++;
                    break;
                case ')':
                    brackets--;
                    break;
                case '"':
                    text = !text;
                    break;
            }
            if(brackets == 0)
            {
                return pos;
            }
            pos++;
        }
        return code.length() - 1;
    }
    
    private static ArrayList<String> splitComma(String code)
    {
        ArrayList<String> list = new ArrayList<>();
        int length = code.length();
        int pos = 0;
        int old = 0;
        int brackets = 0;
        char c;
        boolean text = false;
        while(pos < length)
        {
            c = code.charAt(pos);
            if(text && c != '"')
            {
                pos++;
                continue;
            }
            
            switch(c)
            {
                case '"':
                    text = !text;
                    break;
                case '(':
                    brackets++;
                    break;
                case ')':
                    brackets--;
                    break;
                case ',':
                    if(brackets == 0)
                    {
                        list.add(code.substring(old, pos));
                        old = pos + 1;
                    }
                    break;
            }
            pos++;
        }
        if(old < pos)
        {
            list.add(code.substring(old, pos));
        }
        return list;
    }
    
    private static void splitFunctions(ArrayList<Code> list, String f, int level, int line)
    {
        sublines++;
        int start = findOpenBracket(f, 0);
        int end = findClosingBracket(f, start);
        if(start >= end - 1)
        {
            if(start == f.length())
            {
                list.add(new Code(level, 0, line, sublines, convertInput(f.substring(0, start))));
                return;
            }
            list.add(new Code(f.substring(0, start), level, 0, line, sublines));
            return;
        }
        ArrayList<String> splitted = splitComma(f.substring(start + 1, end));
        if(start == f.length())
        {
            list.add(new Code(level, 0, line, sublines, convertInput(f.substring(0, start))));
        }
        else
        {
            list.add(new Code(f.substring(0, start), level, splitted.size(), line, sublines));
        }
        splitted.forEach(s -> splitFunctions(list, s, level, line));
    }
     
    private static Object convertInput(String s)
    {
        if(s == null)
        {
            return null;
        }
        s = s.trim();
        if(s.startsWith("\"") && s.endsWith("\""))
        {
            if(s.length() <= 1)
            {
                return "\"";
            }
            return s.substring(1, s.length() - 1);
        }
        else if(s.equals("true"))
        {
            return true;
        }
        else if(s.equals("false"))
        {
            return false;
        }
        else if(s.equals("null"))
        {
            return null;
        }
        try
        {
            return new Fraction(Long.parseLong(s));
        }
        catch(NumberFormatException ex)
        {
            return new Variable(s);
        }
    }
}
